/* the main window 
 *
 * $Id: g2main.c,v 1.2 2004/06/26 20:50:15 miq Exp $
 *
 */

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "g2main.h"
#include "g2mesg.h"
#include "g2status.h"
#include "g2map.h"
#include "g2menu.h"
#include "g2text.h"
#include "g2equip.h"
#include "hack.h"

#define VERSION "0.5.4"

static void g2_exit(GtkWidget *app, gpointer data);
static void g2_show_about(GtkWidget * app, gpointer data);

enum {
    META_BIT = 1 << 7,
    CTRL_BITS = 0x1f
};

GtkWidget *mainWin;
GtkWidget *mapWin;
GtkWidget *messageWin;
GtkWidget *equipWin;
GtkWidget *statusWin;
GtkWidget *mainRows;
GtkWidget *topRow;

#if GTK_CHECK_VERSION(2,4,0)

typedef struct {
	gchar *name;
	gchar *label;
	gchar *tooltip;
	gchar *accel;
	gchar command;
} G2ActionEntry;

/* Normal items */
static GtkActionEntry entries[] = {
	{ "FileMenu", NULL, "_File" },
	{ "EditMenu", NULL, "_Edit" },
	{ "ApparelMenu", NULL, "Appa_rel" },
	{ "ActionMenu", NULL, "_Action" },
	{ "SpecialMenu", NULL, "_Special" },
	{ "MoveMenu", NULL, "Move" },
  	{ "HelpMenu", NULL, "_Help" },
  	{ "Quit", GTK_STOCK_QUIT, "_Quit", NULL, "Quit Gtk2hack", G_CALLBACK(g2_exit) },
  	{ "Ride", NULL, "Ride", NULL, NULL, G_CALLBACK(doride) },
  	{ "About", NULL, "_About", NULL, "About Gtk2hack", G_CALLBACK(g2_show_about) },
};

/* XXX perhaps add more commands, tooltips and memnonics; check for sanity with other options */
static G2ActionEntry commands[] = {
	{ "Version", "_Version", "Show Nethack version", "v", 'v' },
	{ "Slash", "Identify a map symbol", NULL, NULL, '/' },
	{ "Backslash", "Discoveries", NULL, NULL, GDK_backslash },
	{ "History", "_History", NULL, "<shift>v", 'V' },
	{ "Ampersand", "Explain command", NULL, NULL, '&' },
	{ "Smaller", "Go up", NULL, NULL, '<' },
	{ "Greater", "Go down", NULL, NULL, '>' },
	{ "Fight", "Attack <direction>", NULL, "F", 'F' },
	{ "Underscore", "Travel", NULL, NULL, '_' },
	{ "Rest", "Rest", NULL, NULL, '.'},
	{ "Apply", "Apply", NULL, "a", 'a'},
	{ "Close", "Close", NULL, "c", 'c'},
	{ "Name", "Name monster", NULL, "<shift>c", 'C'},
	{ "Drop", "Drop", NULL, "d", 'd' },
	{ "DropMany", "Drop many", NULL, "<shift>d", 'D' },
	{ "Kick", "Kick", NULL, "<control>d", CTRL_BITS & GDK_d },
	{ "Eat", "Eat", NULL, "e", 'e' },
	{ "Engrave", "Engrave", NULL, "<shift>e", 'E' },
	{ "Fire", "Fire", NULL, "f", 'f' },
	{ "Inventory", "Inventory", NULL, "i", 'i' },
	{ "Open", "Open", NULL, "o", 'o' },
	{ "Pay", "Pay", NULL, "p", 'p' },
	{ "PutOn", "Put on accessory", NULL, "<shift>p", 'P' },
	{ "Quaff", "Quaff", NULL, "q", 'q' },
	{ "Quiver", "Put in quiver", NULL, "<shift>q", 'Q' },
	{ "Read", "Read", NULL, "r", 'r' },
	{ "Remove", "Remove accessory", NULL, "<shift>r", 'R' },
	{ "Search", "Search", NULL, "s", 's' },
	{ "Throw", "Throw/Shoot", NULL, "t", 't'},
	{ "TakeOff", "Take off armor", NULL, "<shift>t", 'T' },
	{ "Teleport", "Teleport", NULL, "<control>t", CTRL_BITS & GDK_t },
	{ "Wield", "Wield weapon", NULL, "w", 'w' },
	{ "Wear", "Wear armor", NULL, "<shift>w", 'W' },
	{ "Exchange", "Exchange weapons", NULL, "x", 'x' },
	{ "Explore", "Explore mode", NULL, "<shift>x", 'X' },
	{ "Zap", "Zap a wand", NULL, "z", 'z' },
	{ "Cast", "Cast a spell", NULL, "<shift>z", 'Z' },
	{ "PickUp", "Pick up", NULL, NULL, ',' },
	{ "Adjust", "Adjust letters", NULL, "<alt>a", META_BIT | GDK_a },
	{ "Chat", "Chat", NULL, "<alt>c", META_BIT | GDK_c },
	{ "Dip", "Dip", NULL, "<alt>d", META_BIT | GDK_d },
	{ "Enhance", "Enhance skills", NULL, "<alt>e", META_BIT | GDK_e },
	{ "Force", "Force", NULL, "<alt>f", META_BIT | GDK_f },
	{ "Invoke", "Invoke", NULL, "<alt>i", META_BIT | GDK_i },
	{ "Jump", "Jump", NULL, "<alt>j", META_BIT | GDK_j },
	{ "Loot", "Loot", NULL, "<alt>l", META_BIT | GDK_l },
	{ "Monster", "Monster action", NULL, "<alt>m", META_BIT | GDK_m },
	{ "NameIndividual", "Name object", NULL, "<alt>n", META_BIT | GDK_n },
	{ "Offer", "Offer", NULL, "<alt>o", META_BIT | GDK_o },
	{ "Pray", "Pray", NULL, "<alt>p", META_BIT | GDK_p },
	{ "Rub", "Rub", NULL, "<alt>r", META_BIT | GDK_r },
	{ "Sit", "Sit", NULL, "<alt>s", META_BIT | GDK_s },
	{ "Turn", "Turn undead", NULL, "<alt>t", META_BIT | GDK_t },
	{ "Untrap", "Untrap", NULL, "<alt>u", META_BIT | GDK_u },
	{ "Wipe", "Wipe", NULL, "<alt>w", META_BIT | GDK_w },
	{ "Edition", "Show build options", NULL, "<alt>v", META_BIT | GDK_v },
	{ "CharacterInfo", "Show character info", NULL, "<control>x", CTRL_BITS & GDK_x },
};

static const char *uiDescription =
	"<ui>"
	"  <menubar name='MainMenu'>"
	"    <menu action='FileMenu'>"
	"      <menuitem action='Save'/>"
	"      <separator/>"
	"      <menuitem action='Quit'/>"
	"    </menu>"
	"    <menu action='EditMenu'>"
	"      <menuitem action='Options'/>"
	"      <menuitem action='Explore'/>"
	"      <separator/>"
	"      <menuitem action='Adjust'/>"
	"      <menuitem action='Inventory'/>"
	"      <separator/>"
	"      <menuitem action='CharacterInfo'/>"
	"      <menuitem action='Enhance'/>"
	"      <menuitem action='Name'/>"
	"      <menuitem action='NameIndividual'/>"
	"      <menuitem action='Backslash'/>"
	"    </menu>"
	"    <menu action='ApparelMenu'>"
	"      <menuitem action='Wield'/>"
	"      <menuitem action='Exchange'/>"
	"      <menuitem action='Quiver'/>"
	"      <separator/>"
	"      <menuitem action='Wear'/>"
	"      <menuitem action='TakeOff'/>"
	"      <menuitem action='PutOn'/>"
	"      <menuitem action='Remove'/>"
	"    </menu>"
	"    <menu action='ActionMenu'>"
	"      <menuitem action='Fight'/>"
	"      <menuitem action='Throw'/>"
	"      <menuitem action='Fire'/>"
	"      <separator/>"
	"      <menuitem action='Force'/>"
	"      <menuitem action='Untrap'/>"
	"      <menuitem action='Kick'/>"
	"      <menuitem action='Apply'/>"
	"      <menuitem action='Engrave'/>"
	"      <menuitem action='Search'/>"
	"      <menuitem action='Eat'/>"
	"      <menuitem action='Open'/>"
	"      <menuitem action='Close'/>"
	"      <separator/>"
	"      <menuitem action='Loot'/>"
	"      <menuitem action='PickUp'/>"
	"      <menuitem action='Drop'/>"
	"      <menuitem action='DropMany'/>"
	"    </menu>"
	"    <menu action='SpecialMenu'>"
	"      <menuitem action='Quaff'/>"
	"      <menuitem action='Rub'/>"
	"      <menuitem action='Dip'/>"
	"      <separator/>"
	"      <menuitem action='Zap'/>"
	"      <menuitem action='Cast'/>"
	"      <menuitem action='Invoke'/>"
	"      <menuitem action='Read'/>"
	"      <menuitem action='Teleport'/>"
	"      <separator/>"
	"      <menuitem action='Offer'/>"
	"      <menuitem action='Pray'/>"
	"      <menuitem action='Wipe'/>"
	"      <menuitem action='Turn'/>"
	"      <menuitem action='Monster'/>"
	"      <separator/>"
	"      <menuitem action='Pay'/>"
	"      <menuitem action='Chat'/>"
	"    </menu>"
	"    <menu action='MoveMenu'>"
	"      <menuitem action='Jump'/>"
	"      <menuitem action='Ride'/>"
	"      <menuitem action='Sit'/>"
	"      <menuitem action='Smaller'/>"
	"      <menuitem action='Greater'/>"
	"      <menuitem action='Underscore'/>"
	"      <menuitem action='Rest'/>"
	"    </menu>"
	"    <menu action='HelpMenu'>"
	"      <menuitem action='About'/>"
	"      <separator/>"
	"      <menuitem action='Version'/>"
	"      <menuitem action='Edition'/>"
	"      <menuitem action='History'/>"
	"      <menuitem action='Help'/>"
	"      <separator/>"
	"      <menuitem action='Slash'/>"
	"      <menuitem action='Ampersand'/>"
	"    </menu>"
	"  </menubar>"
	"</ui>";


static void command_activated(GtkWidget *action, gpointer command)
{
	keyBuffer = g_slist_append(keyBuffer, GINT_TO_POINTER(command));
}

static void add_new_action_to_group(GtkActionGroup *group, const gchar *name,
		const gchar *label, const gchar *tooltip, const gchar *accel,
		const gchar command, const gchar *stock_id)
{
	GtkAction* action;

	action = gtk_action_new(name, label, tooltip, stock_id);
	g_signal_connect(G_OBJECT(action), "activate", G_CALLBACK(command_activated),
			GINT_TO_POINTER((gint) command));
	gtk_action_group_add_action_with_accel(group, action, accel);
}
#endif

void g2_exit_windows(const gchar* string)
{
    g_print("%s\n", string);
    gtk_widget_destroy(g2_get_main_window());
    exit(0);
}

static void g2_exit(GtkWidget * app, gpointer data)
{
    gint response;
    GtkWidget *saveDialog;
    
    saveDialog = gtk_message_dialog_new(GTK_WINDOW(mainWin), GTK_DIALOG_MODAL,
            GTK_MESSAGE_QUESTION, GTK_BUTTONS_YES_NO, "Do you want to save the "
            "current game?\nIf not your current game will be lost!");
	gtk_dialog_set_default_response (GTK_DIALOG(saveDialog), GTK_RESPONSE_YES);
	response = gtk_dialog_run(GTK_DIALOG(saveDialog));
    if (response == GTK_RESPONSE_YES) {
        dosave0();
    }
    g2_exit_windows("Exiting...\n");
}

static void g2_show_about(GtkWidget * app, gpointer data)
{
    gint response;
    GtkWidget *aboutDialog;
	GtkWidget *vbox;
	GtkWidget *label;
    
    aboutDialog = gtk_dialog_new_with_buttons("About Gtk2Hack", 
			GTK_WINDOW(mainWin), GTK_DIALOG_MODAL | GTK_DIALOG_NO_SEPARATOR,
			GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE, NULL);
	vbox = gtk_vbox_new(FALSE, 5);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 10);
	label = gtk_label_new(NULL);
	gtk_label_set_markup(GTK_LABEL(label), "<b><big>Gtk2Hack " VERSION "</big></b>");
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_CENTER);
	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 10);
	label = gtk_label_new("Gtk2hack was developed by Mihael Vrbanec.\n"
			"Artwork is borrowed from the Qt, X11 and Gnome window ports.\n\n"
			"Thanks for trying Gtk2Hack.");
	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 2);
	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(aboutDialog)->vbox), vbox);
	gtk_widget_show_all(aboutDialog);
    response = gtk_dialog_run(GTK_DIALOG(aboutDialog));
	gtk_widget_destroy(aboutDialog);
}

static gboolean g2_move_keys_with_numpad(guint keyval, gint* key)
{
    switch (keyval) {
    case GDK_Right:
    case GDK_rightarrow:
    case GDK_KP_Right:    
    case GDK_KP_6:
        *key = '6';
        break;
    case GDK_Left:
    case GDK_leftarrow:
    case GDK_KP_Left:    
    case GDK_KP_4:
        *key = '4';
        break;
    case GDK_Up:
    case GDK_uparrow:
    case GDK_KP_Up:    
    case GDK_KP_8:
        *key = '8';
        break;
    case GDK_Down:
    case GDK_downarrow:
    case GDK_KP_Down:    
    case GDK_KP_2:
        *key = '2';
        break;
    case GDK_Home:
    case GDK_KP_Home:    
    case GDK_KP_7:
        *key = '7';
        break;
    case GDK_End:
    case GDK_KP_End:    
    case GDK_KP_1:
        *key = '1';
        break;
    case GDK_Page_Down:
    case GDK_KP_Page_Down:    
    case GDK_KP_3:
        *key = '3';
        break;
    case GDK_Page_Up:
    case GDK_KP_Page_Up:    
    case GDK_KP_9:
        *key = '9';
        break;
    default:
        return FALSE;
        break;
    }
    return TRUE;
}

static gboolean g2_key_is_valid_input(guint keyval)
{
    switch (keyval) {
	case GDK_Return:
	case GDK_KP_Enter:
	case GDK_Escape:
	case GDK_BackSpace:
		return TRUE;
		break;
    default:
        return FALSE;
    }
}

static gboolean g2_main_key_press_event(GtkWidget *widget, GdkEventKey *event,
                                          gpointer data)
{
    gint key;
	
	key = event->keyval;
	/* convert movement keys */
    if (iflags.num_pad && g2_move_keys_with_numpad(event->keyval, &key)) {
		keyBuffer = g_slist_append(keyBuffer, GINT_TO_POINTER(key));
	} else if (g_ascii_isprint(key) || g2_key_is_valid_input(key)) {
		if (event->state & GDK_MOD1_MASK) {
			/*g_print("modifier <ALT> pressed\n");*/
			key = key | META_BIT;
		} else if (event->state & GDK_CONTROL_MASK) {
			/*g_print("modifier <CONTROL> pressed\n");*/
			/* do not handle some menu accelerators which are not needed by nethack */
			switch (key) {
			case GDK_q:
				return FALSE;
			default:
				break;
			}
			key = CTRL_BITS & key;
		}
		keyBuffer = g_slist_append(keyBuffer, GINT_TO_POINTER(key));
	}
	return TRUE;
}

#if GTK_CHECK_VERSION(2,4,0)
static GtkWidget *create_menubar()
{
	GtkWidget *menubar;
	GtkActionGroup *actionGroup;
	GtkUIManager *uiManager;
	GtkAccelGroup *accelGroup;
	GError* error;
	gint i;

	actionGroup = gtk_action_group_new("MenuActions");
	gtk_action_group_add_actions(actionGroup, entries, G_N_ELEMENTS(entries), mainWin);
	for (i = 0; i < G_N_ELEMENTS(commands); i++) {
		add_new_action_to_group(actionGroup, commands[i].name, commands[i].label,
				commands[i].tooltip, commands[i].accel, commands[i].command, NULL);
	}
	/* special-cases for entries with stock ids */
	add_new_action_to_group(actionGroup, "Save", "Save game", NULL,	"<shift>s",
			GDK_S, GTK_STOCK_SAVE);
	add_new_action_to_group(actionGroup, "Options", "Options", NULL, "<shift>o",
			GDK_O, GTK_STOCK_PREFERENCES);
	add_new_action_to_group(actionGroup, "Help", "Show Help Menu", NULL, NULL,
			GDK_question, GTK_STOCK_HELP);
	
	uiManager = gtk_ui_manager_new();
	gtk_ui_manager_insert_action_group(uiManager, actionGroup, 0);

	accelGroup = gtk_ui_manager_get_accel_group(uiManager);
	gtk_window_add_accel_group(GTK_WINDOW(mainWin), accelGroup);

	error = NULL;
	if (!gtk_ui_manager_add_ui_from_string(uiManager, uiDescription, -1, &error)) {
		g_message("building menus failed: %s", error->message);
		g_error_free (error);
		exit(EXIT_FAILURE);
  	}

	menubar = gtk_ui_manager_get_widget(uiManager, "/MainMenu");
	return menubar;
}
#endif

void g2_init_main_window(int *argcp, char **argv)
{
    GtkWidget *menubar;
	int i = 0;
    
    g_print("Gtk2Hack %s\n", VERSION);
    for (i = 0; i < *argcp; i++) {
        g_print("Arg[%d]:%s\n", i, argv[i]);
    }
    /* TODO: learn how to drop privileges when running as suid or sgid */
    gtk_init(argcp, &argv);

    mainWin = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title(GTK_WINDOW(mainWin), "Gtk2Hack " VERSION);
    g_signal_connect(G_OBJECT(mainWin), "delete_event",
                     G_CALLBACK(g2_exit), NULL);

    mainRows = gtk_vbox_new(FALSE, 0);
    gtk_container_set_border_width(GTK_CONTAINER(mainRows), 0);

    topRow = gtk_hbox_new(FALSE, 1);

    mapWin = g2_map_new();
    messageWin = g2_message_new();
	equipWin = g2_equipment_new();
    statusWin = g2_status_new();
    
#if GTK_CHECK_VERSION(2,4,0)
	g_print("Creating gtk2.4-style menubar!\n");	
	menubar = create_menubar();
	gtk_box_pack_start(GTK_BOX(mainRows), menubar, FALSE, FALSE, 0);
#else
	g_print("No pre-gtk2.4 menubar implemented (yet)!\n");
#endif
    gtk_box_pack_start(GTK_BOX(topRow), messageWin, TRUE, TRUE, 2);
    gtk_box_pack_start(GTK_BOX(topRow), equipWin, FALSE, FALSE, 2);
    gtk_box_pack_start(GTK_BOX(topRow), statusWin, FALSE, FALSE, 2);
    gtk_box_pack_start(GTK_BOX(mainRows), topRow, FALSE, FALSE, 2);
    gtk_box_pack_start(GTK_BOX(mainRows), mapWin, FALSE, FALSE, 0);
    g_signal_connect(G_OBJECT(mainWin), "key_press_event",
                     G_CALLBACK(g2_main_key_press_event), NULL);

	gtk_container_add(GTK_CONTAINER(mainWin), mainRows);
    gtk_widget_show_all(mainWin);
}

GtkWidget *g2_get_main_window()
{
    return mainWin;
}

GtkWidget *g2_get_equipment_window()
{
    return equipWin;
}

gint g2_create_window(gint type)
{
    GtkWidget* win;

    switch (type) {
    case NHW_MAP:
        win = mapWin;
        break;
    case NHW_MESSAGE:
        win = messageWin;
        break;
    case NHW_STATUS:
        win = statusWin;
        break;
    case NHW_MENU:
        win = g2_menu_new();
        break;
    case NHW_TEXT:
        win = g2_text_new();
        break;
    default:
        g_print("create unknown window=%d\n", type);
    }

    return (int) win;
}
